/* Macro description */
  /*------------------------------------------------------------------*
   | The documentation and code below is supplied by the Methodology Center, PSU.             
   |              
   *------------------------------------------------------------------*/
                                                                                      
  /*------------------------------------------------------------------*
   | MACRO NAME  : %FHLMLLR
   | SHORT DESC  : Fit a Functional Hierarchical Linear Model for intensive longitudinal data
   *------------------------------------------------------------------*
   | CREATED BY  : Runze LI              (12/04/2009 16:15)
   *------------------------------------------------------------------*
   |
   |
   |  Example: Suppose the dataset simulated_data.sas7bdat has the following variables: 
   |		subj: subjects identification variable; 
   |		t:    measurement time for each observation, different subjects may have different measurement times;
   |		y:    the dependent variable;
   |		x0:   the intercept variable which equals to 1 for all observations;  
   |		x1:   a binary covariate;
   |		x2:   a continuous covariate.   
   |	
   |	Some observations in this data set are as follows:
   |	
   |
   |	subj      Y          t        X0    X1        X2    
   |	-------------------------------------------------
   |	1      0.74397    0.00591     1     0    -1.71718
   |	1      0.20471    0.01772     1     1     0.59689
   |	1     -0.63517    0.02559     1     1    -0.21331
   |	1      2.08357    0.02953     1     1     1.10735
   |	2      0.49824    0.00591     1     1     1.55535
   |	2      0.20461    0.01378     1     1     0.80863
   |	2     -1.05567    0.02165     1     0     0.60380
   |		...	...	...	...	...
   |
   |  To fit a Functional Hierarchical Linear Model(FHLM) in which y is the dependent variable, 
   |	x0, x1, x2 are the covariates with fixed effect, and x1 and x2 are also covariates with random effect, 
   |	but the coefficient(for both random and fixed effects) of x0, x1 and x2 may vary over time,   
   |	we could call the Macro as follows:
   |
   |	%FHLMLLR(  	
   |		mydata = simulated_data,  
   |		id = subj,  		
   |		time = t, 		
   |		dep = y,    		
   |		tcov = x0 x1 x2,  	  
   |		random = x1 x2,  	  
   |		range = 0 1    
   |		perct = 0.2    
   |		);
   |
   |	where cov_knots determines the number of B-spline basis functions used to approximate the corresponding 
   |	coefficeint function, see User's guide for detailed description. 
   |
   |  References:  1) User's guide 
   |               2) Li, R., Root, T. L.,  et al. (2006). A Local Linear Estimation Procedure for Functional Multilevel
   | 				Modeling. Models for intensive longitudinal data. T. A. Walls and J. L. Schafer. New York, 
   | 				NY, US, Oxford University Press. : 63-83.
   |
   *------------------------------------------------------------------*
   |The Macro proceeds as follows: 
   |
   |1) organise the data set into _id_t_dep_cov and _time _t_order 
   |2) using PROC IML and _time,  
   |		a) calculate t_0^i -- the ith center time --  i=1,2,...,N
   |		b) calculate h(t_0^i)-- bandwidth -- based on perct 
   |		c) calculate weight based on t_0^i and h(t_0^i) 
   |		d) output (t_0^i, h(t_0^i)) and (t-t_0^i, w(t)) to two datasets 
   |3) combine _id_t_dep_cov with (t-t_0^i, w(t)) and organize them into qttype2 
   |4) use PROC MIXED to fit the ith local model and save the useful outputs 
   |5) combine outputs into a data set for plot, and other presentations
   |
   *------------------------------------------------------------------*
   | OPERATING SYSTEM COMPATIBILITY
   |
   | UNIX SAS v8   :   
   | UNIX SAS v9   :
   | MVS SAS v8    :
   | MVS SAS v9    :
   | PC SAS v8     :	Yes
   | PC SAS v9     :	Yes
   *------------------------------------------------------------------*
   | Copyright 2009 The Methodology Center, Penn State University.
   |
   | This program is free software; you can redistribute it and/or
   | modify it under the terms of the GNU General Public License as
   | published by the Free Software Foundation; either version 2 of
   | the License, or (at your option) any later version.
   |
   | This program is distributed in the hope that it will be useful,
   | but WITHOUT ANY WARRANTY; without even the implied warranty of
   | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   | General Public License for more details.
   *------------------------------------------------------------------*/

%MACRO FHLMLLR(
		mydata = ,  
		id = ,  			/* the varaible indicating subjects' identification */
		time = , 			/* the varaible contains the measurement times */
		dep = ,    	/*  Y (outcome) */
		tcov = ,  	/* Intercept term should be specified, fixed effects */  
		random = ,  	/* random effects */  
		range = ,		/* range of time */
		perct = 0.3,  		/* percentage of data points for each local model */
		N = 100,				 /* determine the number of fineness of plots valid only when Plot=yes */
		outfilename = 				/* file name for plot_data.csv */
	);

DATA _id_t_dep_cov;
	SET &mydata;
	_mis=0;
RUN;

DATA _id_t_dep_cov;  /* clean the dataset, delete observations with missing covariate*/
	SET _id_t_dep_cov;
	ARRAY a_tcov[*] &dep &time &tcov &random;
      DO i=1 TO DIM(a_tcov);
          IF a_tcov[i]<=. THEN _mis=_mis+1;
      END;
RUN;

DATA _id_t_dep_cov;
	SET _id_t_dep_cov;
	WHERE _mis=0; 
	KEEP &id &time &dep &tcov &random; 
RUN;

DATA _t_orig;
	SET _id_t_dep_cov;
	KEEP &time;
RUN;

PROC SORT DATA=_t_orig OUT=_t_sorted;  /*sorted time variable, needed for bandwidth, weight calculation */ 
	BY &time;
RUN;

/* create null data sets to combine ith output (i=1,...,N) */
DATA est; 
DATA cov;
DATA pre;
DATA t0_h0;


	%LET t_dir = %sysfunc(getoption(work));

	DATA _NULL_;   /* output the generated SAS  code to a temporary file */
    	FILE "&t_dir.\tmpfile.txt";
      	PUT "PROC MIXED COVTEST DATA = _model  METHOD=REML";
		PUT "noclprint noitprint noinfo;";
		PUT "CLASS &id;"; 
       	PUT "MODEL _y0 = " @;

		j = COUNT(COMPBL("&tcov"), " ")+1;
		DO i = 1 TO j;
			tmp = CAT(" _wt*" ,SCAN("&tcov", i));
			PUT tmp @; 
		END;
		DO i = 1 TO j;
			tmp = CAT(" _wt*_t*" ,SCAN("&tcov", i));
			PUT tmp @; 
		END;
		PUT " / NOINT S;";

		IF ( COMPBL("&random") NE '') THEN DO; 
		j = COUNT(COMPBL("&random"), " ")+1; /* COMPBL("&random") */
		PUT "RANDOM  " @; 
		DO i = 1 TO j;
			tmp = CAT(" _wt*" ,SCAN("&random", i));
			PUT tmp @; 
		END;
		*DO i = 1 TO j;
			*tmp = CAT(" _wt*_t*" ,SCAN("&random", i));
			*PUT tmp @; 
		*END;
		PUT " /subject=&id solution;";
		END; /* End of IF */

		PUT "ods output CovParms=_cov SolutionF=_est SolutionR=_pre;";
		PUT "RUN;";
	RUN;

%DO i=1 %TO &N %by 1;
	%Local_i(i = &i, N = &N, perct = &perct); /*generate local data set */

	DATA _model;
		MERGE _id_t_dep_cov _t_wt_i;
		DROP &time;
	RUN;

	DATA _model;
		SET _model;
		WHERE _wt>0;
		_y0 = _wt*&dep;
	RUN;

	 ODS LISTING CLOSE;
  		/* Next we run proc mixed to estimate the parameter using
  		locally weighted data. This corresponds to Step 5 in the
  		proposed algorithm.*/
	%include "&t_dir.\tmpfile.txt";
	ODS LISTING;

DATA _est;
	SET _est;
	_ith = &i;
RUN; 

DATA _cov;
	SET _cov;
	_ith = &i;
RUN; 

DATA _pre;
	SET _pre;
	_ith = &i;
RUN; 

DATA est;
	SET est _est;
RUN;

DATA cov;
	SET cov _cov;
RUN;

DATA pre;
	SET pre _pre;
RUN;

DATA t0_h0;
	SET t0_h0 _t0_h0_i;
RUN;
%END;  /* end of %DO i = ... */

DATA est;
	SET est;
	WHERE _ith>.;
RUN;

DATA cov;
	SET cov;
	WHERE _ith>.;
RUN;

DATA pre;
	SET pre;
	WHERE _ith>.;
RUN;

DATA t0_h0;
	SET t0_h0;
	WHERE _ith>.;
RUN;

%Local_linear_plot(  		
		tcov = &tcov,		
		random = &random  	  
	);


PROC DATASETS;
	DELETE _: ;
RUN; 

%LET t_dir = %sysfunc(getoption(work));

%IF %nrbquote(&outfilename) =  %THEN %DO; 
	%LET outfilename = &t_dir.\plot_data.csv;
%END;

PROC EXPORT DATA= WORK.Plot_data 
            OUTFILE= "&outfilename" 
            DBMS=CSV REPLACE;
RUN;

%MEND FHLMLLR;


%MACRO Local_i( /* generate the ith set of weights, and re-centered time */
		i = 2,   	
		N = 50,       /* total number of grid time points */
		perct = 0.3  /* percent of total observations around the centered time, to be used in the local model*/
	);

  PROC IML;
	USE _t_orig;
	READ ALL INTO mat_t_orig;  

 	USE _t_sorted;
	read ALL INTO mat_t_sorted;

	LL =nrow(mat_t_sorted); /* number of observations */

	low_idx = 1;  	/* round(LL*(&perct)/2, 1); replace 0.1 with &perct/2 */
	upp_idx = LL; 	/* round(LL*(1-(&perct)/2), 1); */

	t_min = mat_t_sorted[low_idx , 1] ; 
	t_max=mat_t_sorted[upp_idx, 1];

	t_range = {&range};   /* 2009-10-08, with range as an input */
	IF ncol(t_range)=2 THEN 
		DO; 
			t_min = max(t_min, t_range[1]);
			t_max = min(t_max, t_range[2]);
		END;

	t0 = t_min + (t_max - t_min)*(&i -1)/(&N -1); /* the ith t0 */ 

	t_t0 = abs(mat_t_sorted - t0);
	CALL sort(t_t0 , {1});

	h_idx = round(LL*(&perct) ,1); /* will replace 0.2 with perct */
	h0 = t_t0[h_idx,1];  /* calculate the bandwidth */

	wt = (1/h0)*(1 - (abs(mat_t_orig - t0)/h0)#(abs(mat_t_orig - t0)/h0)) ; /* calculate weight */
	wt[loc(wt<0),] = 0;  /* delete negative weights */
	wt = sqrt(wt);  /* follow Li's SAS Code */

	mat_t0_h0 = j(1,3); 
	mat_t0_h0[1,1] = t0;
	mat_t0_h0[1,2] = h0;
	mat_t0_h0[1,3] = (&i);

	CREATE _t0_h0_i FROM mat_t0_h0[colname=( ("_t0" || "_h0")||("_ith") ) ]; /* denser time points */
	APPEND FROM mat_t0_h0;

	mat_t_wt = (mat_t_orig - t0) || (wt);

	CREATE _t_wt_i FROM mat_t_wt[colname=( "_t" || "_wt") ]; /* denser time points */
	APPEND FROM mat_t_wt;

  QUIT; 
%MEND Local_i;


%MACRO Local_linear_plot(  		
		tcov = int N_A AR,		
		random = N_A AR AT  	  
	);

	DATA _NULL_;
		n_f = COUNT(COMPBL("&tcov"), " ")+1;

		IF ( COMPBL("&random") NE '') THEN  
			n_r = COUNT(COMPBL("&random"), " ")+1; /* COMPBL("&random") */
		ELSE 
			n_r = 0;

		CALL SYMPUT('n_f', n_f);
		CALL SYMPUT('n_r', n_r);
	 RUN;


	PROC SQL;
		CREATE TABLE plot_data_fix AS
		SELECT A.*, B._t0  
		FROM est A, t0_h0 B 
		WHERE A._ith = B._ith;

	DATA plot_data_fix;
		SET plot_data_fix;
		est_u = estimate + 1.96*stderr;
		est_l = estimate - 1.96*stderr;
	RUN;

	PROC SQL;
		CREATE TABLE plot_data_random AS
		SELECT A.*, B._t0  
		FROM cov A, t0_h0 B 
		WHERE A._ith = B._ith;

	DATA plot_data_random;
		SET plot_data_random;
		est_u = estimate + 1.96*stderr;
		est_l = estimate - 1.96*stderr;
	RUN;

	/* set-up for plots: sumbol 1 - symbol 4 */
		SYMBOL1 
			COLOR=green
			VALUE=point
			INTERPOL=join
        		LINE=2
        		WIDTH=3;
		SYMBOL2 
      		COLOR=red
			VALUE=point
			INTERPOL=join
			LINE=1
	        	WIDTH=3;
		SYMBOL3 
			COLOR=green
			VALUE=point
			INTERPOL=join
			LINE=2
        		WIDTH=3;
		SYMBOL4 
			COLOR=blue
			VALUE=point
			INTERPOL=join
			LINE= 3
        		WIDTH=2;

/* for variance estimates */
	DATA _tmp;
		SET Plot_data_random;
		WHERE COUNT(covparm, "residual", 'i')>0 ; 
		KEEP estimate est_u est_l _t0;
	RUN;

	TITLE1 "Residual Curve and 95% Confidence Band";
	AXIS  
		LABEL=('Residual')
	      WIDTH=3;
		LEGEND LABEL =none value=('Lower Bound' 'Estimated Curve' 'Upper Bound') POSITION=(bottom left inside) MODE=share DOWN = 2; 

	PROC GPLOT DATA=_tmp;
		PLOT (est_l estimate est_U)* _t0 / overlay legend = legend vaxis=axis;
	RUN;

%LET prefix = variance_;	/* for plot data*/
	DATA plot_data;
		SET _tmp; 
		&prefix.est = estimate;
		&prefix.est_l = est_l;
		&prefix.est_u = est_u;
	KEEP  _t0 &prefix.est &prefix.est_l &prefix.est_u;
	RUN;			

/* for fixed effects (cov) */
%LET prefix = fix_;	 /* for plot data*/

	%DO i=1 %TO &n_f;
		%LET var_name = %SCAN(&tcov, &i);
		DATA _tmp;
			SET plot_data_fix;
			WHERE (COUNT(effect, compress("&var_name"), 'i')>0 AND COUNT(effect, "_t", 'i')=0 ); 
			KEEP effect estimate est_u est_l _t0;
		RUN;

		TITLE1 "Coefficient Curve and 95% Confidence Band for Covariate: &var_name";
		AXIS  
			LABEL=('Coefficient' justify=right "of &var_name")
		      WIDTH=3;
		LEGEND LABEL =none value=('Lower Bound' 'Estimated Curve' 'Upper Bound') POSITION=(bottom left inside) MODE=share DOWN = 2; 

		PROC GPLOT DATA=_tmp;
			PLOT (est_l estimate est_U)* _t0 / overlay legend = legend vaxis=axis;
		RUN;

		PROC SQL;
			CREATE TABLE _plot_data AS
			SELECT A.*, B.estimate, B.est_l, B.est_u FROM plot_data A, _tmp B
			    WHERE A._t0=B._t0;
		QUIT;

		DATA plot_data;
			SET _plot_data;
			&prefix&var_name._est = estimate;
			&prefix&var_name._est_l = est_l;
			&prefix&var_name._est_u = est_u;
			DROP estimate est_l est_u ;
		RUN;			
	%END;

/* for random effects (cov) */
 %IF &n_r >0 %THEN %DO;
	%LET prefix = rnd_;	/* for plot data*/

	%DO i=1 %TO &n_r;
		%LET var_name = %SCAN(&random, &i);

		DATA _tmp;
			SET plot_data_random;
			WHERE (COUNT(covparm, compress("&var_name"), 'i')>0 AND COUNT(covparm, "_t", 'i')=0 ); 
			KEEP estimate est_u est_l _t0;
		RUN;

		TITLE1 "Vairance Curve and 95% Confidence Band for: &var_name";
		AXIS  
			LABEL=('Variance' justify=right "of &var_name")
		      WIDTH=3;

		LEGEND LABEL =none value=('Lower Bound' 'Estimated Curve' 'Upper Bound') POSITION=(bottom left inside) MODE=share DOWN = 2; 

		PROC GPLOT DATA=_tmp;
			PLOT (est_l estimate est_U)* _t0 / overlay legend = legend vaxis=axis;
		RUN;

		PROC SQL;
			CREATE TABLE _plot_data AS
			SELECT A.*, B.estimate, B.est_l, B.est_u FROM plot_data A, _tmp B
			    WHERE A._t0=B._t0;
		QUIT;

		DATA plot_data;
			SET _plot_data;
			&prefix&var_name._est = estimate;
			&prefix&var_name._est_l = est_l;
			&prefix&var_name._est_u = est_u;
			DROP estimate est_l est_u;
		RUN;	
	%END;
%END;

	TITLE1;
	SYMBOL1; 
	SYMBOL2; 
	SYMBOL3; 
	SYMBOL4; 
	AXIS;
	LEGEND;

%MEND Local_linear_plot;

